from enum import StrEnum
from typing import Optional

import typer
from anystore.cli import ErrorHandler
from anystore.io import (
    FORMAT_CSV,
    FORMAT_JSON,
    ModelWriter,
    Writer,
    smart_stream_models,
)
from ftmq.io import smart_read_proxies, smart_write_proxies
from rich.console import Console
from typing_extensions import Annotated

from ftm_geocode import __version__, logic
from ftm_geocode.cache import get_cache
from ftm_geocode.geocode import GEOCODERS, geocode_line, geocode_proxy
from ftm_geocode.io import FORMAT_FTM, LatLonRow, PostalRow
from ftm_geocode.logging import configure_logging, get_logger
from ftm_geocode.model import POSTAL_KEYS, GeocodingResult, get_address
from ftm_geocode.nuts import get_nuts, get_proxy_nuts
from ftm_geocode.settings import Settings

settings = Settings()
cli = typer.Typer(no_args_is_help=True)
cli_cache = typer.Typer()
cli.add_typer(cli_cache, name="cache")
console = Console(stderr=True)

log = get_logger(__name__)


class Formats(StrEnum):
    ftm = FORMAT_FTM
    json = FORMAT_JSON
    csv = FORMAT_CSV


class IOFormats(StrEnum):
    json = FORMAT_JSON
    csv = FORMAT_CSV


class Opts:
    IN = typer.Option("-", "-i", help="Input uri (file, http, s3...)")
    OUT = typer.Option("-", "-o", help="Output uri (file, http, s3...)")
    FORMATS = typer.Option(Formats.ftm)
    IOFORMATS = typer.Option(Formats.json)
    GEOCODERS = typer.Option(settings.geocoders, "--geocoders", "-g")
    APPLY_NUTS = typer.Option(False, help="Add EU nuts codes")


@cli.callback(invoke_without_command=True)
def cli_store(
    version: Annotated[Optional[bool], typer.Option(..., help="Show version")] = False,
):
    if version:
        print(__version__)
        raise typer.Exit()
    configure_logging()


@cli.command()
def config():
    """Show current configuration"""
    with ErrorHandler():
        console.print(settings)


@cli.command()
def format_line(
    input_uri: str = Opts.IN,
    input_format: IOFormats = Opts.IOFORMATS,
    output_uri: str = Opts.OUT,
    output_format: IOFormats = Opts.IOFORMATS,
):
    """
    Get formatted lines via libpostal parsing from csv or json input stream with
    1 or more fields:
        - "original_line": address line
        - "country" (optional): country or iso code - good to know for libpostal
        - "language" (optional): language or iso code - good to know for libpostal
        - all other columns will be passed through to the result

    Example:
        ftmgeo format-line -i addresses.csv --input-format csv -o s3://my_bucket/addresses.json
    """
    with ErrorHandler():
        if not settings.libpostal:
            raise Exception("Please install and activate libpostal")
        with ModelWriter(output_uri, output_format=output_format) as writer:
            for row in smart_stream_models(input_uri, PostalRow, input_format):
                row = logic.format_line(row)
                writer.write(row)


@cli.command()
def parse_components(
    input_uri: str = Opts.IN,
    input_format: IOFormats = Opts.IOFORMATS,
    output_uri: str = Opts.OUT,
    output_format: IOFormats = Opts.IOFORMATS,
):
    """
    Get components parsed from libpostal from csv or json input stream with 1 or
    more fields:
        - "original_line": address line
        - "country" (optional): country or iso code - good to know for libpostal
        - "language" (optional): language or iso code - good to know for libpostal
        - all other columns will be passed through to the result

    Example:
        cat data.json | ftmgeo parse-components --output-format csv > data.csv
    """
    with ErrorHandler():
        if not settings.libpostal:
            raise Exception("Please install and activate libpostal")
        rows = smart_stream_models(input_uri, PostalRow, input_format)
        row = next(rows)
        keys = row.model_dump().keys()
        fieldnames = list(set(keys) | set(POSTAL_KEYS))
        with Writer(
            output_uri, output_format=output_format, fieldnames=fieldnames
        ) as writer:
            row = logic.parse_components(row)
            writer.write(row)
            for row in rows:
                row = logic.parse_components(row)
                writer.write(row)


@cli.command()
def map_entities(
    input_uri: str = Opts.IN,
    input_format: IOFormats = Opts.IOFORMATS,
    output_uri: str = Opts.OUT,
):
    """
    Map csv/json input stream to FollowTheMoney Address proxies, requires
    libpostal

    Required input field: `original_line`
    """
    with ErrorHandler():
        if not settings.libpostal:
            raise Exception("Please install and activate libpostal")
        rows = smart_stream_models(input_uri, PostalRow, input_format)
        addresses = (
            get_address(
                r.original_line,
                address_id=r.address_id,
                country=r.country,
                language=r.language,
            )
            for r in rows
        )
        proxies = (a.to_proxy() for a in addresses)
        smart_write_proxies(output_uri, proxies)


@cli.command()
def geocode(
    input_uri: str = Opts.IN,
    input_format: Formats = Opts.FORMATS,
    output_uri: str = Opts.OUT,
    output_format: Formats = Opts.FORMATS,
    geocoders: list[GEOCODERS] = Opts.GEOCODERS,
    use_cache: Annotated[bool, typer.Option(help="Use cache database")] = True,
    cache_only: Annotated[bool, typer.Option(help="Only use cache database")] = False,
    rewrite_ids: Annotated[
        bool, typer.Option(help="Rewrite `Address` entity ids to canonized id")
    ] = True,
    apply_nuts: Annotated[bool, typer.Option(help="Add EU nuts codes")] = False,
):
    """
    Geocode ftm entities or csv input to given output format using different
    geocoders. When using csv input, these columns must be used:

    Columns:
        - "original_line": address line
        - "country" (optional): country or iso code - good to know for libpostal
        - "language" (optional): language or iso code - good to know for libpostal
        - all other columns will be passed through to the result

    Example:
        ftmgeo geocode -i s3://my_bucket/entities.ftm.json > entities.geocoded.ftm.json
    """
    with ErrorHandler():
        if input_format == Formats.ftm:
            proxies = smart_read_proxies(input_uri)
            results = (
                geocode_proxy(
                    geocoders,
                    p,
                    use_cache=use_cache,
                    cache_only=cache_only,
                    apply_nuts=apply_nuts,
                    output_format=output_format,
                    rewrite_ids=rewrite_ids,
                )
                for p in proxies
            )
        else:
            tasks = smart_stream_models(input_uri, PostalRow, input_format)
            results = (
                geocode_line(
                    geocoders,
                    t.original_line,
                    use_cache=use_cache,
                    cache_only=cache_only,
                    country=t.country,
                    apply_nuts=apply_nuts,
                )
                for t in tasks
            )
        out_format = FORMAT_CSV if output_format == FORMAT_CSV else FORMAT_JSON
        with Writer(output_uri, output_format=out_format) as writer:
            for res in results:
                if res is not None:
                    if output_format == FORMAT_FTM:
                        res = res.to_proxy()
                        res = res.to_dict()
                    else:
                        res = res.model_dump(mode="json")
                writer.write(res)


@cli.command()
def apply_nuts(
    input_uri: str = Opts.IN,
    input_format: Formats = Opts.FORMATS,
    output_uri: str = Opts.OUT,
    output_format: IOFormats = Opts.IOFORMATS,
):
    """
    Apply EU NUTS codes to input stream

    For ftm input, only Address entities with longitude and latitude properties
    will be considered.

    For csv or json input, use these fields:
        - "lat": Latitude
        - "lon": Longitude
        - all other fields will be passed through to the result
    """
    with ErrorHandler():
        if input_format == FORMAT_FTM:
            with ModelWriter(output_uri, output_format=output_format) as writer:
                for proxy in smart_read_proxies(input_uri):
                    nuts = get_proxy_nuts(proxy)
                    if nuts is not None:
                        writer.write(nuts)
        else:
            with Writer(output_uri, output_format=output_format) as writer:
                for row in smart_stream_models(input_uri, LatLonRow, input_format):
                    nuts = get_nuts(row.lon, row.lat)
                    if nuts is not None:
                        data = row.model_dump()
                        data.update(nuts.model_dump())
                        writer.write(data)


@cli_cache.command("iterate")
def cache_iterate(
    output_uri: str = Opts.OUT,
    output_format: Formats = Opts.FORMATS,
    apply_nuts: bool = Opts.APPLY_NUTS,
):
    """
    Export cached addresses to csv or ftm entities
    """
    with ErrorHandler():
        cache = get_cache()
        results = cache.iterate_values()
        if apply_nuts:
            results = (r.apply_nuts() for r in results)
        if output_format in (FORMAT_CSV, FORMAT_JSON):
            with ModelWriter(output_uri, output_format=output_format) as writer:
                for res in results:
                    writer.write(res)
        else:
            proxies = (r.to_proxy() for r in results)
            smart_write_proxies(output_uri, proxies)


@cli_cache.command("populate")
def cache_populate(
    input_uri: str = Opts.IN,
    input_format: IOFormats = Opts.IOFORMATS,
    apply_nuts: bool = Opts.APPLY_NUTS,
):
    """
    Populate cache from csv or json input with these fields:\n
        address_id: str\n
        canonical_id: str\n
        original_line: str\n
        result_line: str\n
        country: str\n
        lat: float\n
        lon: float\n
        geocoder: str\n
        geocoder_place_id: str | None = None\n
        geocoder_raw: str | None = None\n
        nuts0_id: str | None = None\n
        nuts1_id: str | None = None\n
        nuts2_id: str | None = None\n
        nuts3_id: str | None = None\n
        ts: datetime | None = None\n
    """
    with ErrorHandler():
        cache = get_cache()
        for row in smart_stream_models(input_uri, GeocodingResult, input_format):
            if apply_nuts:
                row.apply_nuts()
            cache.put(row.address_id, row)



from typing import Generator, Iterator, Literal, TypeAlias

from ftmq.util import get_country_code
from pydantic import BaseModel, ConfigDict

from ftm_geocode.logging import get_logger
from ftm_geocode.model import GeocodingResult, PostalContext

log = get_logger(__name__)


FORMAT_FTM = "ftm"
Formats: TypeAlias = Literal["csv", "json", "ftm"]


class PostalRow(BaseModel):
    original_line: str
    formatted_line: str | None = None
    country: str | None = None
    language: str | None = None

    model_config = ConfigDict(extra="allow")

    @property
    def ctx(self) -> PostalContext:
        return {"country": get_country_code(self.country), "language": self.language}


class LatLonRow(BaseModel):
    lat: float
    lon: float

    model_config = ConfigDict(extra="allow")


PostalRows: TypeAlias = Generator[PostalRow, None, None]
LatLonRows: TypeAlias = Generator[LatLonRow, None, None]
GeocodingResults: TypeAlias = (
    Generator[GeocodingResult, None, None] | Iterator[GeocodingResult]
)



from ftm_geocode.io import PostalRow
from ftm_geocode.model import get_address, get_components


def format_line(row: PostalRow) -> PostalRow:
    address = get_address(row.original_line, **row.ctx)
    row.formatted_line = address.get_formatted_line()
    return row


def parse_components(row: PostalRow) -> dict[str, str | None]:
    res = get_components(row.original_line, **row.ctx)
    res.update(row.model_dump())
    return res



from enum import StrEnum
from pathlib import Path

from anystore.model import StoreModel
from anystore.settings import Settings as Anystore
from geopy.geocoders import SERVICE_TO_GEOCODER
from pydantic_settings import BaseSettings, SettingsConfigDict

from ftm_geocode import __version__

anystore = Anystore()

NUTS = Path(__file__).parent.parent / "data" / "NUTS_RG_01M_2021_4326.shp.zip"
GEOCODERS = StrEnum("Geocoders", ((k, k) for k in SERVICE_TO_GEOCODER.keys()))


class Settings(BaseSettings):
    """
    `ftm-geocode` settings management using
    [pydantic-settings](https://docs.pydantic.dev/latest/concepts/pydantic_settings/)

    Note:
        All settings can be set via environment variables in uppercase,
        prepending `FTMGEO_` (except for those with a given prefix)

    """

    model_config = SettingsConfigDict(
        env_prefix="ftmgeo_",
        env_nested_delimiter="__",
        nested_model_default_partial_update=True,
    )

    user_agent: str = f"ftm-geocode v{__version__}"
    """User-Agent string to use for geocoding services"""

    default_timeout: int = 10
    """Geocoder timeout"""

    min_delay_seconds: float = 0.5
    """Minimum delay between geocoding requests"""

    max_retries: int = 5
    """Maximum retries for geocoding"""

    cache: StoreModel = StoreModel(uri=anystore.uri)
    """Cache uri (using anystore)"""

    nuts_data: Path = NUTS
    """Location for nuts shapefile data"""

    geocoders: list[GEOCODERS] = [GEOCODERS.nominatim]
    """Default geocoders to use (in order)"""

    libpostal: bool = False
    """Activate libpostal (requires additional install)"""



from functools import cache

from anystore.store import BaseStore, get_store

from ftm_geocode.logging import get_logger
from ftm_geocode.settings import Settings
from ftm_geocode.util import make_address_id

log = get_logger(__name__)
settings = Settings()


def make_cache_key(value, **kwargs) -> str | None:
    if kwargs.get("use_cache") is False:
        return
    return make_address_id(value, **kwargs)


@cache
def get_cache() -> BaseStore:
    from ftm_geocode.model import GeocodingResult

    kwargs = settings.cache.model_dump()
    kwargs["model"] = GeocodingResult
    kwargs["store_none_values"] = False
    return get_store(**kwargs)



from typing import Any, Generator, Iterable
from unicodedata import normalize as _unormalize

from banal import ensure_list
from followthemoney.types import registry
from followthemoney.util import make_entity_id
from ftmq.util import get_country_code, get_country_name
from nomenklatura.entity import CE
from normality import collapse_spaces
from normality import normalize as _normalize
from rigour.addresses import normalize_address


def make_address_id(line: str, country: str | None = None, **kwargs) -> str:
    value = make_entity_id(normalize_address(line))
    assert value, f"Invalid address line for id: {line}"
    ccode = get_country_code(country)
    if ccode is not None:
        value = f"{ccode}-{value}"
    return f"addr-{value}"


def get_first(value: str | Iterable[Any] | None, default: Any | None = None) -> Any:
    value = ensure_list(value)
    if value:
        return value[0]
    return default


def clean_country_codes(values: Iterable[str] | str | None) -> set[str]:
    codes = set()
    for value in ensure_list(values):
        code = get_country_code(value)
        if code is not None:
            codes.add(code)
    return codes


def clean_country_names(values: Iterable[str] | str | None) -> set[str]:
    names = set()
    for value in ensure_list(values):
        name = get_country_name(value)
        if name is not None:
            names.add(name)
    return names


def normalize(value: str) -> str:
    return _unormalize("NFC", collapse_spaces(value))


def normalize_google(value: str) -> str:
    # Google error: "One of the input parameters contains a non-UTF-8 string"
    return ", ".join(_normalize(v, lowercase=False) for v in value.split(","))


def get_proxy_addresses(proxy: CE) -> Generator[str, None, None]:
    if proxy.schema.is_a("Address"):
        yield proxy.caption
    else:
        for value in proxy.get_type_values(registry.address):
            yield value


def apply_address(proxy: CE, address: CE, rewrite_id: bool | None = True) -> CE:
    if proxy.schema.is_a("Address"):
        if rewrite_id:
            proxy.id = address.id
        else:
            address.id = proxy.id
        return proxy.merge(address)
    proxy.add("addressEntity", address.id)  # FIXME delete old reference?
    proxy.add("address", address.caption)
    proxy.add("country", address.get("country"))
    return proxy



import os
from datetime import datetime
from typing import Any, Generator, TypedDict

import geopy.geocoders
from anystore import anycache
from banal import clean_dict
from ftmq.util import ensure_proxy
from geopy.adapters import AdapterHTTPError
from geopy.exc import GeocoderQueryError, GeocoderServiceError
from geopy.extra.rate_limiter import RateLimiter
from geopy.geocoders import get_geocoder_for_service
from nomenklatura.entity import CE
from normality import collapse_spaces

from ftm_geocode.cache import get_cache, make_cache_key
from ftm_geocode.io import FORMAT_FTM, Formats
from ftm_geocode.logging import get_logger
from ftm_geocode.model import Address, GeocodingResult, get_canonical_id
from ftm_geocode.settings import GEOCODERS, Settings
from ftm_geocode.util import (
    apply_address,
    get_country_name,
    get_proxy_addresses,
    normalize,
    normalize_google,
)

settings = Settings()

geopy.geocoders.options.default_user_agent = settings.user_agent
geopy.geocoders.options.default_timeout = settings.default_timeout

log = get_logger(__name__)


class GeocodingContext(TypedDict):
    country: str | None
    language: str | None


class Geocoder:
    SETTINGS = {
        GEOCODERS.nominatim: {
            "config": {
                "domain": os.environ.get("FTMGEO_NOMINATIM_DOMAIN"),
            },
            "params": lambda **ctx: {"country_codes": ctx.get("country")},
        },
        GEOCODERS.googlev3: {
            "config": {
                "api_key": os.environ.get("FTMGEO_GOOGLE_API_KEY"),
            },
            "params": lambda **ctx: {"region": ctx.get("country")},
            "query": lambda query, **ctx: normalize_google(query),
        },
        GEOCODERS.arcgis: {
            "params": lambda **ctx: {"out_fields": "*"},
            "query": lambda query, **ctx: ", ".join(
                (query, get_country_name(ctx.get("country")) or "")
            ),
        },
    }

    def __init__(self, geocoder: GEOCODERS):
        self._settings = self.SETTINGS.get(geocoder, {})
        config = clean_dict(self._settings.get("config", {}))
        self.geocoder = get_geocoder_for_service(geocoder.value)(**config)

    def get_params(self, **ctx: GeocodingContext) -> dict[str, Any]:
        func = self._settings.get("params", lambda **ctx: {})
        return clean_dict(func(**ctx))

    def get_query(self, query: str, **ctx: GeocodingContext) -> str:
        func = self._settings.get("query", lambda query, **ctx: normalize(query))
        return func(query, **ctx)


@anycache(
    store=get_cache(),
    key_func=lambda _, v, **kwargs: make_cache_key(v, **kwargs),
    model=GeocodingResult,
)
def _geocode(
    geocoder: GEOCODERS,
    value: str,
    use_cache: bool | None = True,
    cache_only: bool | None = False,
    **ctx: GeocodingContext,
) -> GeocodingResult | None:
    if cache_only:
        return
    geolocator = Geocoder(geocoder)
    value = geolocator.get_query(value, **ctx)
    geocoding_params = geolocator.get_params(**ctx)
    geocode = RateLimiter(
        geolocator.geocoder.geocode,
        min_delay_seconds=settings.min_delay_seconds,
        max_retries=settings.max_retries,
    )

    try:
        result = geocode(value, **geocoding_params)
    except (AdapterHTTPError, GeocoderQueryError, GeocoderServiceError) as e:
        result = None
        log.error(
            f"{e}: {e.message} `{value}`",
            geocoder=geocoder.value,
            **geocoding_params,
        )

    if result is not None:
        log.info(
            f"Geocoder hit: `{value}`",
            geocoder=geocoder.value,
            **geocoding_params,
        )
        address = Address.from_string(result.address, **ctx)
        geocoder_place_id = result.raw.get("place_id")
        if geocoder_place_id:
            address_id = get_canonical_id(geocoder, geocoder_place_id)
        else:
            address_id = address.get_id()
        result = GeocodingResult(
            cache_key=make_cache_key(value, **ctx),
            address_id=address_id,
            original_line=value,
            result_line=result.address,
            country=address.get_country(),
            lat=result.latitude,
            lon=result.longitude,
            geocoder=geocoder.value,
            geocoder_place_id=geocoder_place_id,
            geocoder_raw=result.raw,
            ts=datetime.now(),
        )
        return result


def geocode_line(
    geocoders: list[GEOCODERS],
    value: str,
    use_cache: bool | None = True,
    cache_only: bool | None = False,
    apply_nuts: bool | None = False,
    **ctx: GeocodingContext,
) -> GeocodingResult | None:
    cleaned_value = collapse_spaces(value)
    if cleaned_value:
        for geocoder in geocoders:
            result = _geocode(
                geocoder,
                cleaned_value,
                use_cache=use_cache,
                cache_only=cache_only,
                **ctx,
            )
            if result is not None:
                if apply_nuts:
                    result.apply_nuts()
                return result

    log.warning(f"No geocoding match found: `{value}`", geocoders=geocoders)


def geocode_proxy(
    geocoder: list[GEOCODERS],
    proxy: CE | dict[str, Any],
    use_cache: bool | None = True,
    cache_only: bool | None = False,
    apply_nuts: bool | None = False,
    output_format: Formats | None = FORMAT_FTM,
    rewrite_ids: bool | None = True,
) -> Generator[CE | GeocodingResult, None, None]:
    proxy = ensure_proxy(proxy)
    if not proxy.schema.is_a("Thing"):
        if output_format == FORMAT_FTM:
            yield proxy
        return

    is_address = proxy.schema.is_a("Address")
    ctx = {"country": proxy.first("country") or ""}
    results = (
        geocode_line(
            geocoder,
            value,
            use_cache=use_cache,
            cache_only=cache_only,
            apply_nuts=apply_nuts,
            **ctx,
        )
        for value in get_proxy_addresses(proxy)
    )
    if output_format == FORMAT_FTM:
        for result in results:
            if result is not None:
                address = Address.from_result(result)
                address = address.to_proxy()
                address.add("country", ctx["country"])
                address.add("region", result.nuts)
                proxy = apply_address(proxy, address, rewrite_id=rewrite_ids)
                if is_address:
                    yield proxy
                else:
                    yield address
        if not is_address:
            yield proxy
    else:
        for result in results:
            if result is not None:
                yield result



from collections import defaultdict
from datetime import datetime
from typing import Any, TypeAlias, TypedDict

import lazy_import
import orjson
from anystore.util import clean_dict
from banal import is_mapping
from followthemoney import model
from followthemoney.util import join_text
from ftmq.types import SDict
from ftmq.util import clean_string, make_proxy
from nomenklatura.entity import CE, CompositeEntity
from normality import collapse_spaces
from pydantic import BaseModel, create_model, field_validator, model_validator
from rigour.addresses import clean_address, format_address_line, normalize_address

from ftm_geocode.cache import make_cache_key
from ftm_geocode.nuts import get_nuts
from ftm_geocode.settings import GEOCODERS, Settings
from ftm_geocode.util import (
    clean_country_codes,
    clean_country_names,
    get_country_code,
    get_first,
)

settings = Settings()
USE_LIBPOSTAL = settings.libpostal


class GeocodingResult(BaseModel):
    cache_key: str
    address_id: str
    original_line: str
    result_line: str
    country: str
    lon: float
    lat: float
    geocoder: str
    geocoder_place_id: str | None = None
    geocoder_raw: dict[str, Any] | None = None
    nuts1_id: str | None = None
    nuts2_id: str | None = None
    nuts3_id: str | None = None
    ts: datetime | None = None

    @property
    def nuts(self) -> tuple[str, str, str] | None:
        if self.nuts1_id:
            return (self.nuts1_id, self.nuts2_id, self.nuts3_id)

    def apply_nuts(self) -> None:
        if not self.nuts1_id or not self.nuts2_id or not self.nuts3_id:
            nuts = get_nuts(self.lon, self.lat)
            if nuts is not None:
                self.nuts1_id = nuts.nuts1_id
                self.nuts2_id = nuts.nuts2_id
                self.nuts3_id = nuts.nuts3_id

    def to_proxy(self) -> CE:
        address = Address.from_result(self)
        proxy = address.to_proxy()
        proxy.add("region", self.nuts)
        return proxy

    @model_validator(mode="before")
    @classmethod
    def make_cache_key(cls, data: SDict) -> SDict:
        data["cache_key"] = make_cache_key(
            data["original_line"], country=data.get("country")
        )
        return data

    @field_validator("geocoder_place_id", mode="before")
    @classmethod
    def to_str(cls, value) -> str | None:
        return clean_string(value)

    @field_validator("geocoder_raw", mode="before")
    @classmethod
    def to_dict(cls, value: Any) -> dict[str, Any]:
        if is_mapping(value):
            return value
        if isinstance(value, (str, bytes)):
            return orjson.loads(value)
        return {}


# https://github.com/openvenues/libpostal#parser-labels
# postal -> ftm
# FIXME extend ftm schema to align with postal output?
MAPPING = (
    ("full", "full"),  # used as dummy prop when USE_LIBPOSTAL=False
    # venue name e.g. "Brooklyn Academy of Music", and building names
    # e.g. "Empire State Building"
    ("house", "remarks"),
    # for category queries like "restaurants", etc.
    ("category", "keywords"),
    # phrases like "in", "near", etc. used after a category phrase to help with parsing
    # queries like "restaurants in Brooklyn"
    ("near", "remarks"),
    # usually refers to the external (street-facing) building number. In some countries
    # this may be a compound, hyphenated number which also includes an apartment number,
    # or a block number (a la Japan), but libpostal will just call it the house_number
    # for simplicity.
    ("house_number", "remarks"),
    # street name(s)
    ("road", "street"),
    # an apartment, unit, office, lot, or other secondary unit designator
    ("unit", "remarks"),
    # expressions indicating a floor number e.g. "3rd Floor", "Ground Floor", etc.
    ("level", "remarks"),
    # numbered/lettered staircase
    ("staircase", "remarks"),
    # numbered/lettered entrance
    ("entrance", "remarks"),
    # post office box: typically found in non-physical (mail-only) addresses
    ("po_box", "postOfficeBox"),
    # postal codes used for mail sorting
    ("postcode", "postalCode"),
    # usually an unofficial neighborhood name like "Harlem", "South Bronx", or
    # "Crown Heights"
    ("suburb", "remarks"),
    # these are usually boroughs or districts within a city that serve some official
    # purpose e.g. "Brooklyn" or "Hackney" or "Bratislava IV"
    ("city_district", "remarks"),
    # any human settlement including cities, towns, villages, hamlets, localities, etc.
    ("city", "city"),
    # named islands e.g. "Maui"
    ("island", "region"),
    # usually a second-level administrative division or county.
    ("state_district", "region"),
    # a first-level administrative division. Scotland, Northern Ireland, Wales, and
    # England in the UK are mapped to "state" as well (convention used in OSM,
    # GeoPlanet, etc.)
    ("state", "state"),
    # informal subdivision of a country without any political status
    ("country_region", "region"),
    # sovereign nations and their dependent territories, anything with an ISO-3166 code.
    ("country", "country"),
    ("country_code", "country"),
    # currently only used for appending “West Indies” after the country name, a pattern
    # frequently used in the English-speaking Caribbean e.g. “Jamaica, West Indies”
    ("world_region", "region"),
)

POSTAL_KEYS = [m[0] for m in MAPPING]

Values = list[str] | None


class PostalContext(TypedDict):
    language: str | None
    country: str | None


class AddressBase(BaseModel):
    def get_country(self) -> str:
        return ";".join(self.country or [])

    def get_first(self, attr, default: Any | None = None) -> str | None:
        return get_first(getattr(self, attr, None), default)

    def get_id(self) -> str:  # serves as cache key
        line = normalize_address(self.get_formatted_line(), latinize=True)
        key = make_cache_key(line, country=self.get_first("country"))
        assert key is not None
        return key

    def to_dict(self) -> dict[str, list[str]]:
        return clean_dict(self.model_dump())


class PostalAddressBase(AddressBase):
    full: Values = None
    country_code: Values = None

    def __init__(self, **data):
        data["country"] = clean_country_names(data.get("country"))
        data["country_code"] = clean_country_codes(data.get("country"))
        super().__init__(**data)

    def get_formatted_line(self) -> str:
        country = self.get_first("country")
        data = {
            "attention": self.get_first("near"),
            "house": join_text(self.get_first("house"), self.get_first("po_box")),
            "house_number": self.get_first("house_number"),
            "road": self.get_first("street") or self.get_first("road"),
            "postcode": self.get_first("postcode"),
            "city": self.get_first("city"),
            "state": self.get_first("state"),
            "country": country,
        }
        return format_address_line(data, country=country)

    def to_dict(self) -> dict[str, str]:
        return clean_dict({k: get_first(v) for k, v in self.model_dump().items()})

    @classmethod
    def from_postal_result(
        cls, input_data: list[tuple[str, str]], **ctx: PostalContext
    ) -> "PostalAddress":
        data = defaultdict(set)
        for value, key in input_data:
            data[key].add(value.title())
        if "country" in ctx:
            data["country"].add(ctx["country"])
        return cls(**data)

    @classmethod
    def from_string(cls, value: str, **ctx: PostalContext) -> "PostalAddress":
        value = clean_address(value)
        if USE_LIBPOSTAL:
            parse_address = lazy_import.lazy_callable("postal.parser.parse_address")
            # postal screams if language or country is None
            ctx = {k: ctx.get(k, "") or "" for k in ("language", "country")}
            result = parse_address(value, **ctx)
        else:
            result = [(value, "full")]
        return cls.from_postal_result(result, **ctx)


PostalAddress: PostalAddressBase = create_model(
    "PostalAddress",
    **{k: (Values, None) for k in POSTAL_KEYS},
    __base__=PostalAddressBase,
)

FtmAddressBase: AddressBase = create_model(
    "FtmAddressBase",
    **{p: (Values, None) for p in model.get("Address").properties},
    __base__=AddressBase,
)


class Address(FtmAddressBase):
    _id: str | None = None
    _postal: PostalAddress | None = None

    def get_id(self) -> str:
        # use place ids to generate ids
        if self._id:
            return self._id
        osmId, googlePlaceId = self.get_first("osmId"), self.get_first("googlePlaceId")
        if osmId:
            return f"addr-osm-{osmId}"
        if googlePlaceId:
            return f"addr-google-{googlePlaceId}"
        return super().get_id()

    def to_proxy(self) -> CE:
        return make_proxy(
            {
                "id": self.get_id(),
                "schema": "Address",
                "properties": clean_dict(self.model_dump()),
            }
        )

    def get_formatted_line(self) -> str:
        country = get_country_code(self.get_first("country"))
        data = {
            "attention": collapse_spaces(
                " ".join((self.get_first("summary", ""), " ".join(self.remarks or [])))
            ),
            "house": self.get_first("postOfficeBox"),
            "road": self.get_first("road")
            or self.get_first("street")
            or self.get_first("full"),
            "postcode": self.get_first("postalCode"),
            "city": self.get_first("city"),
            "state": self.get_first("state"),
        }
        return format_address_line(data, country=country)

    @classmethod
    def from_postal(cls, input_data: PostalAddress, **ctx: PostalContext) -> "Address":
        mapping = dict(MAPPING)
        data = defaultdict(set)
        data["country"].add(ctx.get("country"))
        for key, values in input_data:
            if key == "country_code":
                key = "country"
            if values is not None:
                data[mapping[key]].update(values)
        data["country"] = clean_country_codes(data["country"])
        instance = cls(**data)
        instance._postal = input_data
        return instance

    @classmethod
    def from_string(cls, value: str, **ctx: PostalContext) -> "Address":
        value = clean_address(value)
        return cls.from_postal(PostalAddress.from_string(value, **ctx))

    @classmethod
    def from_result(cls, result: GeocodingResult) -> "Address":
        ctx = {"country": result.country}
        address = cls.from_postal(PostalAddress.from_string(result.result_line, **ctx))
        address.full = [result.result_line]
        address.longitude = [str(result.lon)]
        address.latitude = [str(result.lat)]
        if result.geocoder == GEOCODERS.nominatim.name:
            address.osmId = [result.geocoder_place_id]
        if result.geocoder == GEOCODERS.google.name:
            address.googlePlaceId = [result.geocoder_place_id]
        return address

    @classmethod
    def from_proxy(cls, proxy: CE) -> "Address":
        data = proxy.to_dict()
        address = cls(**data["properties"])
        address._id = proxy.id
        return address


AddressInput: TypeAlias = str | Address | PostalAddress | CE | GeocodingResult


def get_address(
    data: AddressInput, address_id: str | None = None, **ctx: PostalContext
) -> Address:
    if isinstance(data, str):
        addr = Address.from_string(data, **ctx)
    elif isinstance(data, PostalAddress):
        addr = Address.from_postal(data, **ctx)
    elif isinstance(data, CompositeEntity):
        addr = Address.from_proxy(data)
    elif isinstance(data, GeocodingResult):
        addr = Address.from_result(data)
    else:
        raise ValueError(f"Invalid input format: {data}")
    if address_id is not None:
        addr._id = address_id
    return addr


def get_components(data: AddressInput, **ctx: PostalContext) -> dict[str, str | None]:
    if isinstance(data, PostalAddress):
        return data.to_dict()
    if isinstance(data, Address):
        return data._postal.to_dict()

    if isinstance(data, str):
        data = PostalAddress.from_string(data, **ctx)
    elif isinstance(data, CompositeEntity):
        data = PostalAddress.from_string(data.caption, **ctx)
    elif isinstance(data, GeocodingResult):
        data = PostalAddress.from_string(GeocodingResult.result_line, **ctx)
    else:
        raise NotImplementedError(data)

    return data.to_dict()


def get_formatted_line(data: AddressInput, **ctx: PostalContext) -> str:
    address = get_address(data, **ctx)
    return address.get_formatted_line()


def get_canonical_id(geocoder: GEOCODERS, place_id: str) -> str:
    if geocoder == GEOCODERS.nominatim:
        return f"addr-osm-{place_id}"
    return f"addr-{geocoder.value}-{place_id}"


def get_coords(data: AddressInput, **ctx: PostalContext) -> tuple[float, float] | None:
    address = get_address(data, **ctx)
    try:
        return float(get_first(address.longitude)), float(get_first(address.latitude))
    except ValueError:
        return None



"""
apply nuts codes to geocoded address

https://ec.europa.eu/eurostat/web/gisco/geodata/reference-data/administrative-units-statistical-units/nuts
https://en.wikipedia.org/wiki/Nomenclature_of_Territorial_Units_for_Statistics
"""

from functools import cache
from typing import Any, Self

import geopandas as gpd
from ftmq.util import get_country_name
from nomenklatura.entity import CE
from pydantic import BaseModel
from shapely.geometry import Point

from ftm_geocode.logging import get_logger
from ftm_geocode.settings import Settings

log = get_logger(__name__)
settings = Settings()


class Nuts(BaseModel):
    level: int
    code: str
    name: str
    country: str
    country_name: str
    path: str

    @classmethod
    def from_code(cls, code: str) -> Self:
        country = code[:2]
        return cls(
            level=len(code) - 2,
            code=code,
            name=get_nuts_name(code),
            country=country,
            country_name=get_country_name(country),
            path=get_nuts_path(code),
        )


class Nuts3(BaseModel):
    nuts1: str
    nuts1_id: str
    nuts2: str
    nuts2_id: str
    nuts3: str
    nuts3_id: str
    country: str
    country_name: str
    path: str

    @classmethod
    def from_code(cls, code: str) -> Self:
        nuts = split_nuts3(code)
        c, n1, n2, n3 = nuts
        return cls(
            nuts1=get_nuts_name(n1),
            nuts1_id=n1,
            nuts2=get_nuts_name(n2),
            nuts2_id=n2,
            nuts3=get_nuts_name(n3),
            nuts3_id=n3,
            country=c,
            country_name=get_country_name(c),
            path="/".join(nuts),
        )


class ProxyNuts(Nuts3):
    entity_id: str


def split_nuts3(code: str) -> tuple[str, str, str, str]:
    # country, nuts1, nuts2, nuts3
    return code[:2], code[:3], code[:4], code[:5]


@cache
def get_nuts_data():
    log.info("Loading nuts shapefile", fp=settings.nuts_data)
    df = gpd.read_file(settings.nuts_data)
    df = df[["LEVL_CODE", "NUTS_ID", "NUTS_NAME", "geometry"]]
    return df


@cache
def get_nuts_names():
    df = get_nuts_data()
    df = df[["NUTS_ID", "NUTS_NAME"]].drop_duplicates().set_index("NUTS_ID")
    return df["NUTS_NAME"].T.to_dict()


def get_nuts_name(code: str) -> str:
    names = get_nuts_names()
    return names[code]


def get_nuts_path(code: str) -> str:
    return "/".join([code[: i + 2] for i in range(len(code) - 1)])


def _get_nuts(lon: float, lat: float) -> Nuts3 | None:
    df = get_nuts_data()
    df = df[df["LEVL_CODE"] == 3]
    point = Point(lon, lat)
    res = df[df.contains(point)].drop_duplicates(subset=("NUTS_ID",))
    if res.empty:
        return
    if len(res) > 1:
        log.error("Invalid nuts lookup result, got %d values instead of 1" % len(res))
        return
    for _, row in res.iterrows():
        return Nuts3.from_code(row["NUTS_ID"])


def get_nuts(lon: Any | None = None, lat: Any | None = None) -> Nuts3 | None:
    try:
        lon, lat = round(float(lon), 6), round(float(lat), 6)
        return _get_nuts(lon, lat)
    except ValueError:
        log.error("Invalid coordinates: (%s, %s)" % (lon, lat))


def get_proxy_nuts(proxy: CE) -> ProxyNuts | None:
    if not proxy.schema.is_a("Address"):
        return
    try:
        lon, lat = proxy.first("longitude"), proxy.first("latitude")
        if lon is not None and lat is not None:
            lon, lat = float(lon), float(lat)
            lon, lat = round(lon, 6), round(lat, 6)  # EU shapefile precision
            nuts = get_nuts(lon, lat)
            if nuts is not None:
                return ProxyNuts(entity_id=proxy.id, **nuts.model_dump())
    except ValueError:
        log.error("Invalid coords", proxy=proxy.to_dict())
        return
